
## main function to run MCMC
mdp <- function(n=NULL, # sample size
                p=NULL, # dimension of responses
                n.iter=1000,   # n MCMC iterations
                n.discard=100, # initial transient
                n.reinit=10000, # reinitialize
                n.batch=50,     # batch size to thin out
                n.predupdate=100, # batch size to update predictives
                n.printallpars=1000, # batch size to print all parameters 
                m.prior=0,      # resample m
                B.prior=0,      # resample B
                verbose=1,      # level of comments
                pxy=0,          # indicator for bivariate predictive dens
                mupxy=0,        # bivariate post dens for mu[n+1]
                exy=0,          # posterior predictive surface
                exvecy=0,
                px=0,           # marginal posterior predictive
                ex=0,           # posterior mean curve
                ix=NULL,        # index for x in pxy, mupxy, exy..
                iy=NULL,        # index for y
                iz=NULL,
                xgrid=NULL,
                nx=NULL,
                ygrid=NULL,
                ny=NULL,
                nz=NULL,        # number of simulations z ~ E(z |x,y,th,dta)s
                s=15,           # d.f. for prior on S
                S.init=NULL,    # initial value and prior mean for S
                q=5,            # d.f. for prior on R
                R=NULL,         # hyperprior parameter R
                B.init=NULL,    # initial value for B,
                cc=5,            # d.f. par c
                C=NULL,         # hyperprior par C
                m.init=NULL,    # initial values for m
                a=NULL,         # hyperpar a
                A=NULL,         # hyperpar A
                alpha=1,        # initial value for total mass alpha
                a0=1,
                b0=1,           # hyperprior pars for alpha ~ IG(a0/2,b0/2)
                k0=NULL,        # initial number of clusters
                Y=NULL,         # data (could be alphanumeric filename)
                mu=NULL)        # initial values (m.l.e?) for mu[i]
{
  options(digits=3)
  
  ## 1. check basic parameters
  if(is.null(Y)){
    cat("\n *** Error: must specify data Y (matrix or filename).\n")
    return(-1)
  }
  if(!is.numeric(Y)){
    dta <- read.table(Y)
    if (is.null(dta)){ ## reading in data did not work
      cat("\n *** Error: could not read data from file",Y,
          "\n Hint: need to include full path.\n")
      return(-1)
    }
    if (verbose>0){
      cat("Read ", nrow(dta), " rows and ", ncol(dta), " columns",
          "from data file ", Y,"\n")
    }
    Y <- as.matrix(dta)
  }
  if(is.null(n))
    n <- nrow(Y)
  else
    if (n != nrow(Y)){
      cat("\n *** Error: nrows(Y) != n. Need to match.\n")
      return(-1)
    }
  if(is.null(p))
    p <- ncol(Y)
  else
    if(p != ncol(Y)){
      cat("\n *** Error: ncol(Y) != p. Need to match.\n")
      return(-1)
    }
  if (n.discard >= n.iter){
    cat("\n *** Error: n.iter <= n.discard. \n")
    return(-1)
  }
  if (n.batch >= n.iter/2){
    cat("\n *** Error: n.iter/2 <= n.batch. \n")
    return(-1)
  }
  if (!is.null(S.init))
    if ( (nrow(S.init) != p) | (ncol(S.init) != p)){
      cat("\n *** Error: dim(S.init) != (p x p).\n")
      return(-1)
    }
  if (!is.null(B.init))
    if ( (nrow(B.init) != p) | (ncol(B.init) != p)){
      cat("\n *** Error: dim(B.init) != (p x p).\n")
      return(-1)
    }
  if (!is.null(C))
    if ( (nrow(C) != p) | (ncol(C) != p)){
      cat("\n *** Error: dim(C) != (p x p).\n")
      return(-1)
    }
  if (!is.null(A))
    if ( (nrow(A) != p) | (ncol(A) != p)){
      cat("\n *** Error: dim(A) != (p x p).\n")
      return(-1)
    }
  if (!is.null(m.init))
    if (length(m.init) != p){
      cat("\n *** Error: length(m) != p.\n")
      return(-1)
    }
  if (!is.null(a))
    if (length(a) != p){
      cat("\n *** Error: length(a) != p.\n")
      return(-1)
    }

  ## 2. default parameters
  mhat <- apply(Y,2,mean)
  V <- var(Y)
  v <- diag(V)
  if(is.null(S.init)){
    S.init <- 0.25*V
    cat(" setting default for S.init = 1/4*var(Y).\n")
  }
  if(is.null(R)){
    R <- 0.25*diag(v)
    cat(" setting default for R = 1/4*diag(v).\n")
  }
  if(is.null(B.init)){
    B.init <- diag(v)
    cat(" setting default for B.init = var(v).\n")
  }
  if(is.null(m.init)){
    m.init <- mhat
    cat(" setting default for m.init = mean(Y).\n")
  }
  if(is.null(C)){
    C <- diag(v)
    cat(" setting default for C = diag(v).\n")
  }
  if(is.null(A)){
    A <- diag(v)
    cat(" setting default for A = diag(v).\n")
  }
  if(is.null(a)){
    a <- mhat
    cat(" setting default for a = mean(Y).\n")
  }
  if(is.null(k0)){
    k0 <- n
    cat(" setting default for k0=n.\n")
  }
  if ( (k0 != 1) & (k0 != n)){
    cat("\n *** Error: k0 not in {1,n}.\n")
    return(-1)
  }

  ## 3. warnings
  if (cc < p+2)
    cat(" *** Warning: should use c > p+2 for good mixing MCMC.\n")
  if (q < p+2)
    cat(" *** Warning: should use q > p+2 for good mixing MCMC.\n")
  if (s < p+2)
    cat(" *** Warning: should use s > p+2 for good mixing MCMC.\n")
    
  ## 3. write init file
  out <- file("init.mdp",open="w")
  cat(" n ",n,
        "\n p",p,
        "\n niter ", n.iter,
        "\n m_prior ", m.prior, "B_prior ", B.prior,
        "ndiscard ", n.discard, "nreinit ", n.reinit,
        "npi ", n.batch, "\n",
        "\n npredupdate ", n.predupdate,
        "\n nprintallpars ", n.printallpars,
        "\n verbose ", verbose,
        "\n pxy ", pxy,
        "\n mupxy ", mupxy,
        "\n exy ", exy,
        "\n exvecy ", exvecy,
        "\n px ", px,
        "\n ex ", ex,
        "\n s ", s,
        "\n S-init \n", format(S.init),
        "\n q ", q,
        "\n R-inv  \n", format(solve(R)),
        "\n B-init \n", format(B.init),
        "\n c ", cc,
        "\n C \n ", format(C),
        "\n m-init ", format(m.init),
        "\n aa ", format(a),
        "\n A-inv \n", format(solve(A)),
        "\n alpha-init ", alpha,
        "\n a0 ", a0, " b0 ", b0,
        "\n k0 ", k0, "\n",
        file=out)
  close(out)

  ## 4. write data file
  out <- file("data.mdp",open="w")
  cat(" Y \n", file=out)
  write(format(t(Y)),ncolumns=p,file=out)
  close(out)

  ## 5. write ex-init etc.
  ## NOTE: C indexes columns from 0,1..; R uses 1,2..
  ##       therefore use 'xi-1' etc
  if(px){ # univariate predictive
    if (is.null(ix))
      ix <- 1
    if ((ix > p) | (ix < 1) ){
      cat(" *** Error: ix not in {1...p}",
          " (coordinate for predictive p(x | dta)).\n")
      return(-1)
    }
    if (is.null(nx)){
      nx <- 25
      cat(" Setting default nx=25 for grid on x for predictive p(x,y | dta)) .\n")
    }
    if (is.null(xgrid))
      xgrid=seq(min(Y[,ix]),max(Y[,ix]),length=nx)
    out <- file("px-init.mdp",open="w")
    cat(" nx ",nx,"\n xi ",ix-1, file=out)
    cat(" xgrid \n", format(xgrid),"\n",file=out)
    cat("\n xlist 0", file=out)
    close(out)
  }
  if(pxy){ # bivariate predictive
    if (is.null(ix))
      ix <- 1
    if (is.null(iy))
      iy <- 2
    if ((ix > p) | (iy > p) | (ix < 1) | (iy < 1)){
      cat(" *** Error: ix, iy not in {1...p}",
          " (coordinate for predictive p(x y | dta)).\n")
      return(-1)
    }
    if (is.null(nx)){
      nx <- 25
      cat(" Setting default nx=25 for grid on x for predictive p(x,y | dta)) .\n")
    }
    if (is.null(ny)){
      ny <- 25
      cat(" Setting default ny=25 for grid on y for predictive p(x,y | dta).\n")
    }
    if (is.null(xgrid))
      xgrid=seq(min(Y[,ix]),max(Y[,ix]),length=nx)
    if (is.null(ygrid))
      ygrid=seq(min(Y[,iy]),max(Y[,iy]),length=ny)
    out <- file("pxy-init.mdp",open="w")
    cat(" nx ",nx,"\n ny ",ny,"\n xi ",ix-1,"\n yi ",iy-1,"\n", file=out)
    cat(" hpd-alpha 0.66 mu 0 \n",
        " lox ", min(xgrid),
        " hix ", max(xgrid), 
        "\n loy ", min(ygrid),
        " hiy ", max(ygrid),
        "\n xlist 0", file=out)
    close(out)
  }
  if(ex){ # univariate expectation
    if (is.null(ix))
      ix <- 1
    if (is.null(iz))
      iz <- 2
    if (is.null(nz))
      nz <- 25
    if ((ix > p) | (iz > p) | (ix < 1) | (iz < 1) | (ix==iz)){
      cat(" *** Error: ix, iz not in {1...p}",
          "(coordinate for predictive E(z | x, dta)).\n")
      return(-1)
    }
    if (is.null(nx)){
      nx <- 25
      cat(" Setting default nx=25 for grid on x for",
          "predictive E(z | x, dta).\n")
    }
    if (is.null(xgrid))
      xgrid=seq(min(Y[,ix]),max(Y[,ix]),length=nx)
    out <- file("ex-init.mdp",open="w")
    cat(" nx ",nx,"\n nz ",nz," xi ",ix-1,"\n zi ",iz-1,"\n", file=out)
    cat(" xgrid \n", format(xgrid),"\n",file=out)
    close(out)
  }
  if(exy){ # bivariate response surface
    if (is.null(ix))
      ix <- 1
    if (is.null(iy))
      iy <- 1
    if (is.null(iz))
      iz <- 2
    if ((ix > p) | (iy>p) | (iz>p) | (ix<1) | (iy<1) | (iz<1) | (ix==iz)){
      cat(" *** Error: ix, iy, iz not in {1...p}",
          "(coord's for predictive E(z | x,y, dta)).\n")
      return(-1)
    }
    if (is.null(nx)){
      nx <- 25
      cat(" Setting default nx=25 for grid on x ",
          " for predictive E(z |x,y,dta).\n")
    }
    if (is.null(ny)){
      ny <- 25
      cat(" Setting default ny=25 for grid on y",
          "for predictive E(z |x,y,dta).\n")
    }
    if (is.null(xgrid))
      xgrid=seq(min(Y[,ix]),max(Y[,ix]),length=nx)
    if (is.null(ygrid))
      xgrid=seq(min(Y[,iy]),max(Y[,iy]),length=ny)
    out <- file("exy-init.mdp",open="w")
    cat(" nx ",nx,"\n ny ",ny,
         "\n xi ",ix-1, "\n yi ",iy-1, "\n zi ",iz-1,"\n", file=out)
    cat(" xgrid \n", format(xgrid),"\n",file=out)
    cat(" ygrid \n", format(ygrid),"\n",file=out)
    close(out)
  }
  
  ## 4. call C executable
  info <- .C("mdp2",package="mdp")
  if (verbose>1){
    cat(" Results are saved in files ***.mdp. \n",
        " Use functions 'mdp.pxy(.), mdp.exy(.) etc. to plot results. \n")
  }

  
  ## 5. return error code
  return(info)
}

    
##################################################################
# functions to make predictive plots
##################################################################

plt.pxy <- function(xlab="X",ylab="Y",
                    xlim=NULL,ylim=NULL,
                    lg=F,    # plot log density
                    img=T)   # indicator for image vs. contour plot
  { ## plot joint predictive
    ## needs saved file 'pxy-*.mdp' from previous call to mdp(.)

    ## print message..
    cat("\n Looking for files 'pxy-{x,y,z}.mdp' from previous call to 'mdp(.)'",
        "\n in the working directory...\n")

    xgrid <- scan("pxy-x.mdp")
    ygrid <- scan("pxy-y.mdp")
    nx <- length(xgrid)
    ny <- length(ygrid)
    z <- matrix( scan("pxy-z.mdp"), byrow=T,ncol=ny)
    cat("Read (", nx, "x", ny,") matrix of pxy(x,y) over grid on",
        "x (from",min(xgrid)," to ",max(xgrid),
        " and y (from",min(ygrid)," to ", max(ygrid),"\n")
    
    if(is.null(xlim))
      xlim <- range(xgrid)
    if(is.null(ylim))
      ylim <- range(ygrid)
    if(lg){ ## take log
      if(min(z)==0) # avoid zero..
        z <- z+0.001
      z <- log(z)
    }

    if (img)
      image(xgrid,ygrid,z,xlab=xlab,ylab=ylab)
    else
      contour(xgrid,ygrid,z,xlab=xlab,ylab=ylab)
  }

plt.ex <- function(xlab="X",zlab="Z",
                   xlim=NULL,ylim=NULL,
                   bty="l", # box type
                   sd=F,    # add pointwise posterior sd
                   sim=F)   # show p( E(z | x,theta, data))
  { # plot marg expectation E(z | x)
    cat("\n Looking for files 'ex-{x,z}.mdp' from previous call to 'mdp(.)'",
        "\n in the working directory...\n")

    ## read in simulation output
    xgrid <- scan("ex-x.mdp")
    nx <- length(xgrid)
    ez <- scan("ex-ez.mdp")
    if (sd | sim){
      Z <- matrix( scan("ex-z.mdp"), byrow=T, ncol=nx)
      s <- sqrt(apply(Z,2,var))  # pointwise sd
    }

    ## sete default pars
    if(is.null(xlim))
      xlim <- range(xgrid)
    if(is.null(ylim)){
      if(sd | sim)
        ylim <- range(c(ez+s,ez-s))
      else
        ylim <- range(ez)
    }

    ## plot
    plot(xgrid,ez,xlab=xlab,ylab=zlab,xlim=xlim,type="l",bty=bty)
    if (sd){ ## add pointwise sd
      lines(xgrid,ez+s,lty=2,type="l")
      lines(xgrid,ez-s,lty=2,type="l")
    }
    if (sim)
      matlines(xgrid,t(Z),col=3,lty=3)
  }

plt.px <- function(xlab="X",ylab="p(X)",
                   xlim=NULL,ylim=NULL,
                   add=F,
                   bty="l", # box type
                   col.sim = 3,
                   nlines.sim = NULL,
                   sd=F,    # add pointwise posterior sd
                   sim=F)   # show p( p(x| theta, data))
  { # plot marg expectation E(z | x)
    cat("\n Looking for files 'px-{x,z}.mdp' from previous call to 'mdp(.)'",
        "\n in the working directory...\n")

    ## read in simulation output
    xgrid <- scan("px-x.mdp")
    nx <- length(xgrid)
    px <- scan("px-p.mdp")
    if (sd | sim){
      Z <- matrix( scan("px-z.mdp"), byrow=T, ncol=nx)
      s <- sqrt(apply(Z,2,var))  # pointwise sd
      if (is.null(nlines.sim))
        nlines.sim <- nrow(Z)
    }

    ## sete default pars
    if(is.null(xlim))
      xlim <- range(xgrid)
    if(is.null(ylim)){
      if(sd)
        ylim <- range(c(px+s,px-s))
      else if (sim)
        ylim <- range(Z)
      else
        ylim <- range(px)
    }

    ## plot
    if (!add)
      plot(xgrid,px,xlab=xlab,ylab=ylab,xlim=xlim,ylim=ylim,type="l",bty=bty)
    else
      lines(xgrid,px,type="l")
      
    if (sd){ ## add pointwise sd
      lines(xgrid,px+s,lty=2,type="l")
      lines(xgrid,px-s,lty=2,type="l")
    }
    if (sim){
      lines(xgrid,px,type="l",lwd=3)
      matlines(xgrid,t(Z[1:nlines.sim,]),col=col.sim,lty=3)
    }
  }

plt.exy <- function(xlab="X",      # axis labels
                    ylab="Y",
                    zlab="Z",
                    plt.contour=T) # indicator for contour plot
  {## plot surface E(z | x,y, dta)
    cat("\n Looking for files 'exy-{x,y,z}.mdp' from previous call to 'mdp(.)'",
        "\n in the working directory...\n")
    
    xgrid <- scan("exy-x.mdp")
    ygrid <- scan("exy-y.mdp")
    ny <- length(ygrid)
    nx <- length(xgrid)

    ez <- matrix(scan("exy-z.mdp"),byrow=T,ncol=ny)
    if(plt.contour)
      contour(xgrid,ygrid,ez,xlab=xlab,ylab=ylab)
    else
      persp(xgrid,ygrid,ez,xlab=xlab,ylab=ylab,zlab=zlab,bty="n")
}
