/* ============================================================= 
   Multivariate MDP model
   
  ============================================================= 
   Model:
   y[i] | pi[i] is N(y[i]; mu[i], V[i])
                   pi[i] = (mu[i], V[i])
   pi[i]        is DP(G0, alpha)
   G0(pi)       is W(1/V; s, 1/s*1/S) N(mu; m,B), s>p+3
                   i.e. E(1/V)=1/S, E(V)=S*s/(s-p-1)
		        use s > p+3 (for 2nd moments)
   S            is W(S; q, R/q), i.e. E(S)=R
   alpah        is G(a0, b0), E(alpha) = a0/b0

Estimation as described in MacEachern & Peter (1994)
Notation:
   Some of the pi[i]'s can be identical. The list of distinct
   pi[i]'s is denoted by pi*[0]..pi*[n_class].


   Details as in Escobar & West */

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#include "mdp2.h"
#include "bayes.h"
#include "rand.h"
#include "nrutil.h"
#include "matrix.h"
#include "vector.h"
#include "rand-nr.h"
#include "mess.h"
#include "interface.h"
#include "predict.h"

#define Ndiscard 50          /* number of discarded iterations */

/* ***********************************************************
   all MDP pars
  *********************************************************** */
struct MDP_PARS mdp;
int p;

/* -----------------------------------------------------------
   Gibbs iterations
  ----------------------------------------------------------- */
int n_iter,             /* max iterations */
  iter,                 /* current iter number */
  n_discard,            /* discard the first n_discard draws */
  n_reinitialize,         /* reinitialize configurations */
  n_predupdate,
  n_pi;                 /* every n_pi update expectations etc. */

int n_printallpars;     /* intervals to print all pars if 
			   verbose > 2 */
int m_prior, B_prior;
int n_class0, *member0; /* initial config */

/* -----------------------------------------------------------
   various output files
  ----------------------------------------------------------- */
FILE *parFile, *muFile, *VFile, *SFile, *meanFile, *memberFile, *BFile;

/* -----------------------------------------------------------
   debugging
   ----------------------------------------------------------- */
int dbg=0;

/* ============================================================= 
   main
   ============================================================= */ 
/* main() */
void mdp2()
{ 
  init();
  init_output();
  pred_init(mdp);
  init_config();
  gibbs(); 
  finish();
}				  

/* ============================================================= 
   functions called from main
   ============================================================= */ 
/* ***********************************************************
   init
   ***********************************************************
   reads in data, initial par values */
void  init()
{
  int i, j;
  double r;

/* -----------------------------------------------------------
   read in sample size and dim
   ----------------------------------------------------------- */
  openIn("init.mdp");
  scanInt("n", &mdp.n_obs);
  messinttab("nobs", mdp.n_obs);
  scanInt("p", &p);
  messint("p ",p);
  mdp.p = p;

/* -----------------------------------------------------------
   allocate memory for pars & data
   -----------------------------------------------------------  */
  mdp.Y = dmatrix(0,mdp.n_obs-1,0,p-1);
  mdp.V = darray_3(0,mdp.n_obs-1);
  mdp.Vt = darray_3(0,mdp.n_obs-1);
  mdp.Vj = darray_3(0,mdp.n_obs-1);
  mdp.V_cd_inv = darray_3(0,mdp.n_obs-1);
  mdp.Vt_cd_inv = darray_3(0,mdp.n_obs-1);
  for(j=0;j<mdp.n_obs;j++){
    mdp.V[j] = dmatrix(0,p-1,0,p-1);
    mdp.Vj[j] = dmatrix(0,p-1,0,p-1);
    mdp.V_cd_inv[j] = dmatrix(0,p-1,0,p-1);
    mdp.Vt[j] = dmatrix(0,p-1,0,p-1);
    mdp.Vt_cd_inv[j] = dmatrix(0,p-1,0,p-1);
  }
  mdp.mu = dmatrix(0,mdp.n_obs-1,0,p-1);
  mdp.mj = dmatrix(0,mdp.n_obs-1,0,p-1);
  mdp.mean = dvector(0,p-1);

  mdp.S = dmatrix(0,p-1,0,p-1);
  mdp.S_cd_inv = dmatrix(0,p-1,0,p-1);
  mdp.S_inv = dmatrix(0,p-1,0,p-1);
  mdp.R_inv = dmatrix(0,p-1,0,p-1);
  mdp.B = dmatrix(0,p-1,0,p-1);
  mdp.B_inv = dmatrix(0,p-1,0,p-1);
  mdp.B_cd_inv = dmatrix(0,p-1,0,p-1);
  mdp.C = dmatrix(0,p-1,0,p-1);

  mdp.aa = dvector(0,p-1);
  mdp.A_inv = dmatrix(0,p-1,0,p-1);

  /* note: count, first & upd allow for one extra class =
     "new class" */
  mdp.count = ivector(0,mdp.n_obs);
  mdp.first = ivector(0,mdp.n_obs);

  mdp.member = ivector(0,mdp.n_obs-1);
  mdp.updated = ivector(0,mdp.n_obs-1);
  mdp.prev = ivector(0,mdp.n_obs-1);
  mdp.next = ivector(0,mdp.n_obs);
  mdp.new = ivector(0,mdp.n_obs-1);

  member0 = ivector(0,mdp.n_obs-1);

/* -----------------------------------------------------------
   read in hyperpars and initial values
/* ----------------------------------------------------------- */

  scanInt("niter",&n_iter);
  scanInt("m_prior", &m_prior);
  scanInt("B_prior", &B_prior);
  scanInt("ndiscard",&n_discard);
  scanInt("nreinit",&n_reinitialize);
  scanInt("npi",&n_pi);
  scanInt("npredupdate",&n_predupdate);
  scanInt("nprintallpars",&n_printallpars);
  scanInt("verbose", &mdp.verbose);
  scanInt("pxy", &mdp.pxy);
  scanInt("mupxy", &mdp.mupxy);
  scanInt("exy", &mdp.exy);
  scanInt("exvecy", &mdp.exvecy);
  scanInt("px", &mdp.px);
  scanInt("ex", &mdp.ex);
  scanInt("s", &mdp.s);

  /* read and invert S */
  scanDoubleMatrix("S-init",mdp.S,p,p);
  d_invert(p,mdp.S,mdp.S_inv);
  d_chol_decomp(p,mdp.S_inv,mdp.S_cd_inv);

  scanInt("q",&mdp.q);
  scanDoubleMatrix("R-inv",mdp.R_inv, p,p);

  /* read and invert B */
  scanDoubleMatrix("B-init",mdp.B,p,p);
  d_invert(p,mdp.B,mdp.B_inv);
  d_chol_decomp(p,mdp.B_inv,mdp.B_cd_inv);

  scanInt("c",&mdp.c);
  scanDoubleMatrix("C",mdp.C, p,p);
  scanDoubleArray("m-init",mdp.mean, p);
  scanDoubleArray("aa", mdp.aa, p);
  scanDoubleMatrix("A-inv", mdp.A_inv, p,p);
  scanDouble("alpha-init",&mdp.alpha);
  scanDouble("a0", &mdp.a0);
  scanDouble("b0", &mdp.b0);
  if (B_prior == 1) mdp.tau = 1.0;


/* -----------------------------------------------------------
   initial config
/* ----------------------------------------------------------- */
  scanInt("k0", &n_class0);
  if (n_class0 == 1)
    for(i=0;i<mdp.n_obs;i++)
      member0[i] = 0;
  else if (n_class0 == mdp.n_obs)
    for(i=0;i<mdp.n_obs;i++)
      member0[i] = i;
  else if (n_class0 == -2){
    for(i=0;i<mdp.n_obs;i++)
      member0[i] = i%2;
    n_class0 = 2;
  }
  else
    scanIntArray("member0", member0, mdp.n_obs);

  closeIn();

/* -----------------------------------------------------------
   read in data
/* ----------------------------------------------------------- */

  openIn("data.mdp");
  scanDoubleMatrix("Y",mdp.Y, mdp.n_obs, p);
  closeIn();
}
/* ***********************************************************
   open output files
/* *********************************************************** */
void init_output()
{
  parFile = openOut("par.mdp");
  muFile = openOut("mu.mdp");
  VFile = openOut("V.mdp");
  SFile = openOut("S.mdp");
  BFile = openOut("B.mdp");
  meanFile = openOut("mean.mdp");
  memberFile = openOut("member.mdp");
  fclose(muFile); 
  fclose(SFile); 
  fclose(meanFile); 
  fclose(memberFile); 
  fclose(VFile);
  fclose(BFile);
  fclose(parFile);
}

/* ***********************************************************
   initial configuration
/* *********************************************************** */
void init_config()
{
  int i, k, ct;

  /* compute inverse and cd of S */
  d_invert(p,mdp.S,mdp.S_inv);
  d_chol_decomp(p,mdp.S_inv,mdp.S_cd_inv);
  
  /* create n empty groups */
  mdp.n_class = 0;
  for (k=0;k<mdp.n_obs;k++){
    mdp.count[k] = 0;
    mdp.first[k] = -1;
    mdp.updated[k] = 0;
    mdp.new[k] = 0;
    sample_V(k);
  }

  /* put obs's in groups */
  for (i = 0; i < mdp.n_obs; i++){
    k = member0[i];
    add_class(i,k);
  }

  /* sample mu's */
  for(k=0;k<mdp.n_obs;k++){
    sample_mu(k);
  }
  check_class();
}

/* *************************************************
   reinitialize_config
 ************************************************* */
void reinitialize_config()
{
  int i;

  /* remove old config */
  for (i=0;i<mdp.n_obs;i++)
    take_out(i);
  init_config();
}

/* ***********************************************************
   Gibbs Sampler
/* *********************************************************** */
void gibbs()
{
  int time, last_init, k;

  last_init = 0;
  for (iter = 0; iter < n_iter; iter++){
    if ( (iter % n_reinitialize == 0) & (iter > 0)){
      reinitialize_config();
      last_init = iter;
    }
    if (mdp.verbose > -1) print_pars(iter);
    sample_S();
    mdp.kmax = mdp.n_class;
    sample_config();
    for(k=0; k<mdp.n_class; k++){
      sample_mu(k);
      sample_V(k);
    } 
    sample_alpha(); 
    if (B_prior) sample_B();
    if (m_prior) sample_m();
    if (iter-last_init > n_discard){
      update_nclass();
      if (iter % n_pi == 0){
	print_pi();
	write_pars(iter);
	write_nclass();
      }
      if (iter % n_predupdate == 0) /* update predictive */
	pred_update(iter, mdp);
    }
  }
}

/* ***********************************************************
   finish
/* *********************************************************** */
void finish()
{
     print_pi();
     pred_finish(mdp);
     if (mdp.pxy==1) make_hpd(mdp);
}		   
  
/* ============================================================= 
   Conditional draws
/* ============================================================= */ 
	  
/* ***********************************************************
   sample config
/* ***********************************************************
   does the Gibbs over configurations
   Pr(y[i] in class j | ...) proportional to
   for j=0,..n_class-1:
      count[j]*MV-N(y[i]; mut[j], Vt[j])
      mut[j] and Vt[j] computed in make_mut_Vt
   for j=n_class (i.e. new class): same expression with 
                                   count[n_class] = alpha
*/
void sample_config()
{
  int l, k, k_new, ind, s1;
  double u,*qv;      /* prob of pi[ind] = pi*[k] */
  int si, j;
  FILE *pltfile;

  /* allocate memory */
  qv = dvector(0,mdp.n_obs);
  if (iter==500) pltfile = openOut("plt.mdp");

  for (ind = 0; ind < mdp.n_obs; ind++){
    u = runif();
    if (mdp.count[mdp.member[ind]] == 1)
      /* toss coin with prob 1/k */
      if (u > 1.0/(1.0*mdp.n_class)) continue;
    if (mdp.member[ind]==-1){
      printf("\n*** clusterless point %d! \n",ind);
      exit(1);
    }
    take_out(ind);

    /* compute prob pi[ind] = pi*[k]       */
    for(l=0;l<mdp.n_class;l++){
      if (mdp.updated[l]==0)
	setup_mu(l,0);
      qv[l] = mdp.count[l] * mvn_pdf(mdp.Y[ind],mdp.mj[l],mdp.Vt_cd_inv[l],p);
    }

    /* Prob of new class */
    /* qv[n_class = alpha * MV-N(y[i],m,V+tau*B)*/
    if (mdp.updated[mdp.n_class]==0)
      setup_mu(mdp.n_class,0);
    qv[mdp.n_class] = mdp.alpha/(mdp.n_class+1.0)*
      mvn_pdf(mdp.Y[ind],mdp.mj[mdp.n_class],
	      mdp.Vt_cd_inv[mdp.n_class],p);

    /* for plot */
    if (iter==500){
      for(j=0;j<mdp.n_class+1;j++)
	fprintf(pltfile,
		" %d %d %5.3e %d %5.3e %5.3e %5.3e %5.3e %5.3e\n",
		ind,j,qv[j],
		mdp.count[j], mdp.mj[j][0],mdp.mj[j][1],
		mdp.Vt[j][0][0],mdp.Vt[j][0][1],mdp.Vt[j][1][1]);
    }

    /* draw the index of the new class for ind */
    multinomial(1,mdp.n_class+1,qv,&k_new);     
    add_class(ind,k_new);

    if (k_new == mdp.n_class){ 
      if (k_new > mdp.kmax) mdp.kmax = k_new;
    }
  }

  if (iter==500) fclose(pltfile);

  /* free allocated memory */
  free_dvector(qv,0,mdp.n_obs);
  }

/* -----------------------------------------------------------
   check class
/* ----------------------------------------------------------- */
/* checks for bogous classes */

void check_class(){
  int i,j,k;

  for(k=0;k<mdp.n_class;k++){
    for(j=0,i=mdp.first[k]; i!=-1; j++, i = mdp.next[i])
      if (j>mdp.count[k])
	printf("\n*** Error: class %d is bogous.\n", k);
  }
}

/* -----------------------------------------------------------
   take_out
/* ----------------------------------------------------------- */
void take_out(int ind)
{
  int k_old,i;
  

  /* take y[ind] out of it's current class */
  k_old = mdp.member[ind];
  mdp.count[k_old] -= 1;
  mdp.updated[k_old] = 0;
  mdp.member[ind] = -1;
  if (mdp.prev[ind] == -1)
    mdp.first[k_old] = mdp.next[ind];
  else mdp.next[mdp.prev[ind]] = mdp.next[ind];
  mdp.next[ind] = -1;
  mdp.prev[ind] = -1;

  /* if class k=k_old is empty */
  if (mdp.count[k_old] == 0){

    /* relabel class k=n_class-1 as k_old */
    mdp.updated[k_old] = mdp.updated[mdp.n_class-1];
    mdp.updated[mdp.n_class-1] = 0;
    mdp.count[k_old] = mdp.count[mdp.n_class-1];
    mdp.count[mdp.n_class-1] = 0;
    mdp.first[k_old] = mdp.first[mdp.n_class-1];
    mdp.first[mdp.n_class-1] = -1;
    mdp.new[k_old] = mdp.new[mdp.n_class-1];

    A_swap_B(mdp.V[k_old],mdp.V[mdp.n_class-1],mdp.p,mdp.p);
    A_swap_B(mdp.Vt[k_old],mdp.Vt[mdp.n_class-1],mdp.p,mdp.p);
    A_swap_B(mdp.Vj[k_old],mdp.Vj[mdp.n_class-1],mdp.p,mdp.p);
    A_swap_B(mdp.V_cd_inv[k_old],mdp.V_cd_inv[mdp.n_class-1],mdp.p,mdp.p);
    A_swap_B(mdp.Vt_cd_inv[k_old],mdp.Vt_cd_inv[mdp.n_class-1],mdp.p,mdp.p);   
    x_swap_y(mdp.mu[k_old],mdp.mu[mdp.n_class-1],mdp.p);
    x_swap_y(mdp.mj[k_old],mdp.mj[mdp.n_class-1],mdp.p);

    /* update member-references to the relabeled class */
    for(i=mdp.first[k_old];i!=-1;i=mdp.next[i])
      mdp.member[i] = k_old;

    /* decrement n-class by one */
    mdp.n_class -= 1;
    
    check_class();

  }
}
/* -----------------------------------------------------------
   add_class
/* -----------------------------------------------------------
   adds obs[ind] to class[k]
*/
void add_class(int ind, int k)
{
  mdp.updated[k] = 0;
  mdp.count[k] += 1;
  if (mdp.count[k] == 1){
    mdp.new[k] = 1;
    mdp.n_class += 1;
  }
  mdp.member[ind] = k;
  mdp.next[ind] = mdp.first[k];
  mdp.first[k] = ind;
  if (mdp.count[k] > 1) mdp.prev[mdp.next[ind]] = ind;
  mdp.prev[ind] = -1;

  check_class();
}
/* -----------------------------------------------------------
   make_ybar
/* ----------------------------------------------------------- */
void make_ybar(int k, double *ybar)
{
  int i;
  double nk;

  nk = 1.0*mdp.count[k];
  x_zero(ybar,p); /* initialize ybar=0 */
  for(i=mdp.first[k];i!=-1;i=mdp.next[i]){
    x_plus_y(ybar,mdp.Y[i],ybar,p);
  }
  if (mdp.count[k] != 0) x_div_r(ybar,nk,ybar,p);
}
/* -----------------------------------------------------------
   make_Sk
/* ----------------------------------------------------------- */
void make_Sk(int k, double **S)
{
  int i;
  double nk, *z;

  /* allocate memory */
  z = dvector(0,p-1);

  nk = 1.0*mdp.count[k];
  A_zero(S,p);
  for(i=mdp.first[k];i!=-1;i=mdp.next[i]){
    x_min_y(mdp.Y[i],mdp.mu[k],z,p);
    A_plus_rxxt(S,1.0,z,S,p);
  }

  free_dvector(z,0,p-1);
}

/* *************************************************
   setup_mu
 ************************************************* */
void setup_mu(int k, int draw)
{
  double **V_inv, *ybar, nk;
  double **Vj, **Vj_inv;


  /* allocate memory for the auxilary arrays */
  ybar = dvector(0,p-1);
  V_inv = dmatrix(0,p-1,0,p-1);
  Vj_inv = dmatrix(0,p-1,0,p-1);
  Vj = dmatrix(0,p-1,0,p-1);

  if (mdp.count[k] == 0){
    B_assgn_A(Vj,mdp.B,mdp.p);
    y_assgn_x(mdp.mj[k],mdp.mean,p);
  }
  else{
    /* get pars for conditional posterior p(mu|...) */
    make_ybar(k, ybar);
    d_invert(p,mdp.V[k], V_inv);
    nk = 1.0*mdp.count[k];
    
    /* compute posterior mean cov matrix for mu[k] */
    nn_bayes(1.0,mdp.B_inv,mdp.mean, 1.0/nk, V_inv, ybar,
	     Vj, Vj_inv, mdp.mj[k], p);
  }

  /* keep for update_pred */
  B_assgn_A(mdp.Vj[k],Vj,mdp.p);

  /* if draw, then sample mu[k] */
  if (draw==1)
    mvnS_rand(p,mdp.mj[k],Vj,mdp.mu[k]);

  /* Vt = (Vj + V[k]) */
  rA_plus_sB(1.0,Vj,1.0,mdp.V[k],mdp.Vt[k],p);
  invcd(p,mdp.Vt[k],mdp.Vt_cd_inv[k]);

  mdp.updated[k] = 1;

  /* free memory */
  free_dvector(ybar, 0,p-1);
  free_dmatrix(V_inv,0,p-1,0,p-1);
  free_dmatrix(Vj_inv,0,p-1,0,p-1);
  free_dmatrix(Vj,0,p-1,0,p-1);
}

/* ***********************************************************
   sample_mu
/* ***********************************************************
   sample mu[k] from N(mu; m(y), T)
   where {m(y),T} = T*(nj*V-inv[k]*ybar[k] + B-inv/tau*m),
                    T-inv = nj*V-inv[k] + B-inv/tau.
*/
void sample_mu(int k)
{

  setup_mu(k,1);
}

/* ***********************************************************
   sample_V
/* *********************************************************** */
/* V[k] from W(1/V[k]; s+n[k], (sS+S[k])^-1),
   where S[k] = sum (y[i]-mu[k])(y[i]-mu[k])'
*/
void sample_V(int k)
{ double **Sk, **S1, **S1_inv_cd,
    **V_cd, **V_inv;

  /* allocate memory */
  V_cd = dmatrix(0,p-1,0,p-1);
  V_inv = dmatrix(0,p-1,0,p-1);
  Sk = dmatrix(0,p-1,0,p-1);
  S1 = dmatrix(0,p-1,0,p-1);
  S1_inv_cd = dmatrix(0,p-1,0,p-1);
  
  /* Sk = sum (y[i]-mu[j])(y[i]-mu[j]) over member[i]=j */
  make_Sk(k,Sk);

  /* S1 = s*S+Sk */
  rA_plus_sB(mdp.s*1.0, mdp.S, 1.0, Sk, S1, p);
  cdinv(p,S1,S1_inv_cd);
  wishart(mdp.s+mdp.count[k], p, S1_inv_cd, V_inv);
  
  /* compute V, V_cd, V_inv_cd etc. */
  d_chol_decomp(p,V_inv, mdp.V_cd_inv[k]);
  d_inv_triang(p,mdp.V_cd_inv[k],V_cd);
  ABt(V_cd,p,p,V_cd,p,p,mdp.V[k]);

  /* error check */
  if (isnan(mdp.V[k][0][0])){
    error("sample_V", "Vt is NaN.", 1);
  }

  /* allocate memory for the auxilary arrays */
  free_dmatrix(V_cd,0,p-1,0,p-1);
  free_dmatrix(V_inv,0,p-1,0,p-1);
  free_dmatrix(S1,0,p-1,0,p-1);
  free_dmatrix(Sk,0,p-1,0,p-1);
  free_dmatrix(S1_inv_cd,0,p-1,0,p-1);
}
    


/* ***********************************************************
   sample S
/* *********************************************************** 
   generates S from W(S; q+k*s, U)
   U = {(R/q)^-1 + s*sum V_inv[k]}^(-1)
*/
void sample_S()
{
  int k;

  double **U_inv, **U_cd, **V_inv, s;

  /* allocate memory for aux arrays */
  V_inv = dmatrix(0,p-1,0,p-1);
  U_inv = dmatrix(0,p-1,0,p-1);
  U_cd = dmatrix(0,p-1,0,p-1);

  /* initialize U_inv = (R/q)^-1 = q/R */
  rA(1.0*mdp.q, mdp.R_inv, U_inv, p, p); /* 1.0*q is double */

  /* loop over all k classes, add up Vbar_inv and mu */
  s = mdp.s*1.0; /* s_double is double ! */
  for(k=0;k<mdp.n_class;k++){
    /* add (s-p-1)*mdp.V_inv[k] */
    d_invert(p, mdp.V[k], V_inv);
    AplusBr(U_inv, V_inv, s, U_inv, p);
  }
  
  /* compute U_cd */
  cdinv(p,U_inv,U_cd);

  /* generate S from Wish(q+s*k, U) */
  wishart(mdp.q+mdp.n_class*mdp.s, p, U_cd, mdp.S);
  invcd(p,mdp.S,mdp.S_cd_inv);

  /* free auxilary arrays */
  free_dmatrix(V_inv,0,p-1,0,p-1);
  free_dmatrix(U_inv,0,p-1,0,p-1);
  free_dmatrix(U_cd,0,p-1,0,p-1);

  return;
}

/* ***********************************************************
   sample m
/* *********************************************************** 
/* m from p(m|..) = N(m; m(mubar), T)
   m(ybar[k]) = T*(B-inv/tau*k*mubar + A-inv*a)
   T-inv = B-inv/tau*k + A-inv
*/
void sample_m()
{
  double *mubar, K;
  int i;

  K = mdp.n_class; /* need double */

  /* allocate memory */
  mubar = dvector(0,p-1);
  
  /* make mubar = 1/K sum mu[k] */
  x_zero(mubar,p); /* initialize mubar=0 */
  for(i=0;i<mdp.n_class;i++){
    x_plus_y(mubar,mdp.mu[i],mubar,p);
  }
  x_div_r(mubar,K,mubar,p);

  /* compute and sample posterior on m */
  nn_bayes_rand(1.0,mdp.A_inv, mdp.aa,
		mdp.tau/K, mdp.B_inv, mubar,
		mdp.mean, p);

  /* free memory */
  free_dvector(mubar,0,p-1);
}

/* ***********************************************************
   draw B
/* *********************************************************** */
/* dummy */
/* sample B from p(B|..) =
   W(1/B; k+c, [cC + 1/tau*sum (mu[k]-m)(mu[k]-m)']^-1)
*/
void sample_B()
{ double **Sk, **S1, **S1_inv_cd,
    **B_cd, *z;
  int k;

  /* allocate memory */
  B_cd = dmatrix(0,p-1,0,p-1);
  Sk = dmatrix(0,p-1,0,p-1);
  S1 = dmatrix(0,p-1,0,p-1);
  S1_inv_cd = dmatrix(0,p-1,0,p-1);
  z = dvector(0,p-1);
  
  /* Sk = sum (mu[k]-m)(mu[k]-m)' */
  A_zero(Sk,p);
  for(k=0;k<mdp.n_class;k++){
    x_min_y(mdp.mu[k],mdp.mean,z,p);
    A_plus_rxxt(Sk,1.0,z,Sk,p);
  }

  /* S1 = s*S+Sk */
  rA_plus_sB(mdp.c*1.0, mdp.C, 1.0, Sk, S1, p);
  cdinv(p,S1,S1_inv_cd);
  wishart(mdp.c+mdp.n_class, p, S1_inv_cd, mdp.B_inv);
  
  /* compute B, B_cd, B_inv_cd etc. */
  d_chol_decomp(p,mdp.B_inv, mdp.B_cd_inv);
  d_inv_triang(p,mdp.B_cd_inv,B_cd);
  ABt(B_cd,p,p,B_cd,p,p,mdp.B);

  /* error check */
  if (isnan(mdp.B[0][0])){
    error("sample_B", "B is NaN.", 1);
  }

  /* allocate memory for the auxilary arrays */
  free_dmatrix(B_cd,0,p-1,0,p-1);
  free_dmatrix(S1,0,p-1,0,p-1);
  free_dmatrix(Sk,0,p-1,0,p-1);
  free_dmatrix(S1_inv_cd,0,p-1,0,p-1);
  free_dvector(z,0,p-1);
}



/* ***********************************************************
   sample alpha
/* *********************************************************** 
   draw alpha and eta from:
   alpha from     pi*Gamma(a0+n_class, b0-log(eta)) +
              (1-pi)*Gamma(a0+n_class-1, b0-log(eta)),
         where pi = (a+n-class-1)/(a0+n_class-1 + n*b0 - n*log(eta)).
   eta   from Beta(alpha+1,n).
*/
void sample_alpha()
{
     double a,  b, pi;       /* parameters in Gamma call */
     double u;

      dbg = 0;
      if (dbg) message("sample alpha");

     /* draw eta */
     mdp.eta = betadev(mdp.alpha+1.0, mdp.n_obs*1.0);

     /* compute pars for gammas for alpha */
     b = mdp.b0 - log(mdp.eta);
     pi = (mdp.a0 + mdp.n_class - 1)/
       (mdp.a0 + mdp.n_class -1 + mdp.n_obs*(mdp.b0 - log(mdp.eta)));
     duniform(1,&u);
     a = (u<pi) ? mdp.a0+mdp.n_class : mdp.a0+mdp.n_class-1;

     /* draw alpha with prob pi from Gamma(a1,b) else Gamma(a2,b) */
     mdp.alpha = gamdev_ab(a,b);

     if (dbg){
       messdouble("alpha: ",mdp.alpha);
     }
     dbg=0;

}

/* ============================================================= 
   printing routines
/* ============================================================= */ 

/* ***********************************************************
   print pi
/* ***********************************************************
   prints current pi's */
void print_pi()
{
     int k, i;
     
     muFile = openAppend("mu.mdp");
     for (k = 0; k < mdp.n_class; k++){
       fprintf(muFile," %4d",iter);
       fprintf(muFile,"\t %4d \t", mdp.count[k]);
       fwriteDoubleArray(muFile, mdp.mu[k], 1, p);
     }
     fprintf(muFile,"\n");
     fclose(muFile);

     VFile = openAppend("V.mdp");
     for (k = 0; k < mdp.n_class; k++){
       fprintf(VFile,"\n %4d",iter);
       fprintf(VFile,"\t %4d\n", mdp.count[k]);
       fwriteDoubleMatrix2(VFile,mdp.V[k], p, p);
     }
     fclose(VFile);
}

/* ***********************************************************
   print pars
/* ***********************************************************
   print current tau,alpha,eta */
void print_pars(int time)
{
  static int linect=20;
  int i,k, ct[3], k1[3];
  
  if (mdp.verbose < 2){
    printf("%d:%d ",iter,mdp.n_class);
    if (iter % 10 == 0) printf("\n");
    return;
  }

  /* print header line if linect == 20 */
  if (linect == 20){
    if (B_prior) 
      printf("\n\n\n %5s %16s %5s ", "iter","k (count1,2,3)  ","alpha");
    else
      printf("\n\n\n %5s %16s %5s %5s ", "iter","k (count1,2,3)  ",
	     "alpha", "tau");
    for(i=0;i<p;i++) printf("mean[%1d](S,B)    ",i);
    if (mdp.verbose){
      for(k=0;(k<3)&(k<mdp.n_class);k++) 
	for(i=0;i<p;i++) printf("mu[%1d][%1d]    ",k,i);
      printf
	("\n---------------------------------------------------- \n");
      linect = 0;
    }
    printf("\n");
  }
  /* find largest 3 classes */
  k1[0] = k1[1] = k1[2] = 0;
  ct[0] = ct[1] = ct[2] = 0;
  for(k=0; k<mdp.n_class; k++){
    if (mdp.count[k] > ct[0]){
      k1[2] = k1[1];
      ct[2] = ct[1];
      k1[1] = k1[1];
      ct[1] = ct[0];
      k1[0] = k;
      ct[0] = mdp.count[k];
    }
    else if (mdp.count[k] > ct[1]){
      k1[2] = k1[1];
      ct[2] = ct[1];
      k1[1] = k;
      ct[1] = mdp.count[k];
    }
    else if (mdp.count[k] > ct[2]){
      k1[2] = k;
      ct[2] = mdp.count[k];
    }
  }

  /* print line */
  if (B_prior)
    printf(  "%5d %3d(%3d %3d %3d) %5.2f ",
	   iter,mdp.n_class,ct[0], ct[1], ct[2],
	   mdp.alpha);
  else
    printf(  "%5d %3d(%3d %3d %3d) %5.2f %5.2f",
	   iter,mdp.n_class,ct[0], ct[1], ct[2],
	   mdp.alpha, mdp.tau);
  for(i=0;i<p;i++) printf("%4.1f(%4.1f,%4.1f) ", mdp.mean[i], 
			  sqrt(mdp.S[i][i]), sqrt(mdp.B[i][i]));
  if (mdp.verbose){
    for(k=0;(k<3)&(k<mdp.n_class);k++){
      if (mdp.new[k1[k]] == (iter-1)) printf("* ");
      else printf("| ");
      for(i=0;i<p;i++) printf("%4.1f(%4.1f) ",
			      mdp.mu[k1[k]][i],sqrt(mdp.V[k1[k]][i][i]));
    }
  }
  printf("\n");
  linect += 1;
}


int print_allpars()
{
  int k;

  messdoublematrix2("B ",mdp.B,p,p);
  messdoublematrix2("S ",mdp.S,p,p);
  messdoublevec("m ",mdp.mean,p);
  for(k=0;k<mdp.n_class;k++){
    messdoublevec("mu[i] ",mdp.mu[k],p);
    messdoublematrix2("V[j] ",mdp.V[k],p,p);
  }
  printf("Note: sampling seq is \t  S,config,mu,V,B,m.\n");
  printf("------------------------------------------------------- \n");
}

/* ***********************************************************
   write pars
/* ***********************************************************
   print current tau,alpha,eta */
void write_pars(int time)
{
  parFile = openAppend("par.mdp");
  fprintf(parFile,"%4d  %6.3f %6.3f %4d\n",
	  iter, mdp.alpha,mdp.eta, mdp.n_class);
  fclose(parFile);
  meanFile = openAppend("mean.mdp");
  fprintf(meanFile,"%4d \t", iter);
  fwriteDoubleArray(meanFile, mdp.mean, 1, p);
  fclose(meanFile);
  SFile = openAppend("S.mdp");
  fprintf(SFile,"%4d \t", iter);
  fwriteDoubleMatrix2(SFile, mdp.S, p,p);
  fclose(SFile);
  if (B_prior){
    BFile = openAppend("B.mdp");
    fprintf(BFile,"%4d \t", iter);
    fwriteDoubleMatrix2(BFile, mdp.B, p,p);
    fclose(BFile);
  }
  memberFile = openAppend("member.mdp");
  fprintf(memberFile,"%4d \t", iter);
  fwriteIntArray(memberFile, mdp.member, 1, mdp.n_obs);
  fclose(memberFile);
}


/* ============================================================= 
  ergodic avges for group numbers
/* ============================================================= */ 

int nclass_initialized=0; /* indicator whether nclass is initialized */
double *nclass; /* accumulate number of clusters */
int max_nclass = 0;

/* ***********************************************************
   update_nclass
/* *********************************************************** */
int update_nclass()
{
  int j;

  if (nclass_initialized == 0){
    nclass = dvector(0,mdp.n_obs);
    for (j=0;j<=mdp.n_obs;j++)
      nclass[j] = 0.0;
    nclass_initialized = 1;
  }

  max_nclass = (mdp.n_class > max_nclass) ? mdp.n_class : max_nclass;
  nclass[mdp.n_class] += 1.0;
}


int write_nclass()
{
  int j;
  FILE *pjfile;

  pjfile = openOut("pj.mdp");
  fwriteDoubleArray(pjfile, nclass, 10, max_nclass);
  fclose(pjfile);
}
